// SPDX-FileCopyrightText: 2025 Gleb Smirnov <glebsmirnov0708@gmail.com>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::future::Future;

use gtk::glib;

pub fn spawn_async_local(f: impl Future<Output = ()> + 'static) {
    glib::MainContext::default().spawn_local(f);
}
